import os
import tempfile
from typing import Optional, List

import inquirer
import requests
import typer

from openapi_client import (
    ProtoVolumeMountRequest,
    ProtoVolumeMountRequestSourceDataset,
    ProtoVolumeMountRequestSourceVolume,
)
from savvihub.api.exceptions import NotFoundAPIException
from savvihub.api.file_object import DownloadableFileObject
from savvihub.api.uploader import Uploader
from savvihub.cli.exceptions import ExitException
from savvihub.cli.typer import Context
from savvihub.cli.utils import find_from_inquirer, parse_dataset


def start_command_callback(start_command: str) -> str:
    if start_command:
        return start_command

    return inquirer.prompt([inquirer.Text(
        'start_command',
        message="Start command",
        default="python main.py",
    )]).get('start_command')


def cluster_name_callback(ctx: Context, cluster_name: str) -> str:
    assert ctx.params['workspace_name']

    clusters = [c for c in ctx.authenticated_client.cluster_list(ctx.params['workspace_name']).clusters
                if c.status == 'connected']
    if cluster_name:
        for cluster in clusters:
            if cluster.name == cluster_name.strip():
                ctx.store['cluster'] = cluster
                return cluster.name
        else:
            raise ExitException(f'Cluster not found: {cluster_name.strip()}')

    elif len(clusters) == 1:
        typer.echo(f'The cluster is automatically set to `{clusters[0].name}{" (SavviHub)" if clusters[0].is_savvihub_managed else f" (Custom)"}`.')
        ctx.store['cluster'] = clusters[0]
        return clusters[0].name

    else:
        selected_cluster = find_from_inquirer(
            clusters,
            lambda x: f'{x.name}{" (SavviHub)" if x.is_savvihub_managed else f" ({x.name})"}',
            "Please choose a cluster"
        )
        ctx.store['cluster'] = selected_cluster

    return selected_cluster.name


def resource_name_callback(ctx: Context, resource_name: str) -> Optional[str]:
    assert ctx.params['workspace_name'] and ctx.store['cluster']

    if ctx.store['cluster'].is_savvihub_managed:
        resources = ctx.authenticated_client.kernel_resource_list(ctx.params['workspace_name']).results
        if resource_name:
            for resource in resources:
                if resource.name == resource_name.strip():
                    ctx.store['processor_type'] = resource.processor_type
                    ctx.store['resource_spec_id'] = resource.id
                    return resource.name
            else:
                raise ExitException(f'Resource not found: {resource_name.strip()}')
        else:
            selected_resource = find_from_inquirer(
                resources,
                lambda x: f'{x.name} ({x.description})',
                "Please choose a resource"
            )
            ctx.store['processor_type'] = selected_resource.processor_type
            ctx.store['resource_spec_id'] = selected_resource.id
            return selected_resource.name

    elif resource_name:
        raise ExitException('--resource option can be set only with a savvihub-managed cluster')

    return None


def processor_type_callback(ctx: Context, processor_type: str) -> str:
    assert ctx.store['cluster']

    if ctx.store['cluster'].is_savvihub_managed:
        return ctx.store['processor_type']

    processor_type = processor_type.upper() if processor_type else None
    if processor_type not in ['CPU', 'GPU']:
        processor_type = find_from_inquirer(
            ['CPU', 'GPU'],
            lambda x: x,
            "Please choose a processor type",
        )
    return processor_type


def cpu_limit_callback(ctx: Context, cpu_limit: float) -> Optional[float]:
    assert ctx.store['cluster']
    if ctx.store['cluster'].is_savvihub_managed:
        return 0

    if cpu_limit <= 0:
        cpu_limit = float(inquirer.prompt([inquirer.Text(
            'question',
            message="CPU limit (the number of vCPUs)",
            default=1.0,
        )]).get('question'))
        if cpu_limit <= 0:
            raise ExitException('Must be greater than 0')

    return cpu_limit


def memory_limit_callback(ctx: Context, memory_limit: float) -> Optional[str]:
    assert ctx.store['cluster']
    if ctx.store['cluster'].is_savvihub_managed:
        return None

    if not memory_limit or memory_limit <= 0:
        memory_limit = float(inquirer.prompt([inquirer.Text(
            'question',
            message="Memory limit in GiB",
            default='4.0',
        )]).get('question'))
        if memory_limit <= 0:
            raise ExitException('Must be greater than 0')

    memory_limit = str(memory_limit) + 'Gi'
    return memory_limit


def gpu_type_callback(ctx: Context, gpu_type: str) -> str:
    assert ctx.store['cluster'] and ctx.params['processor_type']
    if ctx.store['cluster'].is_savvihub_managed or ctx.params['processor_type'] != 'GPU':
        return 'Empty'

    if gpu_type:
        return gpu_type

    nodes = ctx.authenticated_client.cluster_node_list(
        ctx.params['workspace_name'], ctx.store['cluster'].name).nodes
    ctx.store['nodes'] = nodes
    gpu_type = find_from_inquirer(
        nodes,
        lambda x: f'{x.gpu_product_name}: ({x.gpu_allocatable - x.gpu_limits}/{x.gpu_allocatable})',
        "Please choose the GPU type"
    ).gpu_product_name
    return gpu_type


def gpu_limit_callback(ctx: Context, gpu_limit: int) -> int:
    assert ctx.store['cluster'] and ctx.params['gpu_type']
    if ctx.store['cluster'].is_savvihub_managed or ctx.params['gpu_type'] == 'Empty':
        return 0

    if gpu_limit <= 0:
        gpu_limit = int(inquirer.prompt([inquirer.Text(
            'question',
            message="GPU limit (the number of GPUs)",
            default=1,
        )]).get('question'))
        if gpu_limit <= 0:
            raise ExitException('Must be greater than 0')
    return gpu_limit


def image_url_callback(ctx: Context, image_url: str) -> str:
    assert ctx.params['workspace_name'] and ctx.params['processor_type']

    images = ctx.authenticated_client.kernel_image_list(ctx.params['workspace_name']).results
    images = [i for i in images if i.processor_type == ctx.params['processor_type']]
    if image_url:
        for image in images:
            if image.image_url == image_url:
                return image_url
        else:
            raise ExitException(f'Image not found: {image_url}')

    return find_from_inquirer(
        images,
        lambda x: f'{x.image_url} ({x.name})',
        "Please choose a kernel image"
    ).image_url


def dataset_mount_callback(ctx: Context, dataset_mounts: List[str]) -> List[str]:
    client = ctx.authenticated_client
    workspace_name = ctx.params['workspace_name']
    ctx.store['dataset_mounts'] = []
    for dataset_mount in dataset_mounts:
        splitted = dataset_mount.split(':')
        if len(splitted) != 2:
            raise ExitException(f'Invalid dataset mount format: {dataset_mount}. '
                                f'You should specify both mount path and dataset name.\n'
                                f'ex) /input/dataset1:mnist@snapshot-3d1e0f91c')

        mount_path, dataset_full_name = splitted
        workspace_name_override, dataset_name, snapshot_name = parse_dataset(dataset_full_name)
        if workspace_name_override:
            workspace_name = workspace_name_override

        dataset = client.dataset_read(workspace_name, dataset_name)
        if snapshot_name != 'latest':
            try:
                client.snapshot_read(dataset.volume_id, snapshot_name)
            except NotFoundAPIException:
                raise ExitException(f'Invalid dataset snapshot: {dataset_full_name}\n'
                                    f'Please check your dataset and snapshot exist in workspace `{workspace_name}`.')

        ctx.store['dataset_mounts'].append(ProtoVolumeMountRequest(
            mount_type='dataset',
            mount_path=mount_path,
            dataset=ProtoVolumeMountRequestSourceDataset(
                dataset_id=dataset.id,
                snapshot_name=snapshot_name,
            ),
        ))

    return dataset_mounts


def volume_file_mount_callback(ctx: Context, volume_file_mounts: List[str]) -> List[str]:
    ctx.store['volume_file_mounts'] = []
    for volume_file_mount in volume_file_mounts:
        splitted = volume_file_mount.split(':')
        if len(splitted) != 2:
            raise ExitException(f'Invalid volume file mount format: {volume_file_mount}. '
                                f'You should specify both mount path and volume file.\n'
                                f'Volume file can be notated with volume id and subpath.'
                                f'ex) /input/model.pt:1#subPath=f72fca375e812/model.pt')

        mount_path, volume_id = splitted
        if '#subPath=' in volume_id:
            volume_id, sub_path = volume_id.split('#subPath=')
        else:
            sub_path = ''

        try:
            volume_id = int(volume_id)
        except ValueError:
            raise ExitException(f'Volume id should be an integer.')

        try:
            ctx.authenticated_client.volume_read(volume_id)
        except NotFoundAPIException:
            raise ExitException(f'Volume not found: {volume_id}')

        ctx.store['volume_file_mounts'].append(ProtoVolumeMountRequest(
            mount_type='volume',
            mount_path=mount_path,
            volume=ProtoVolumeMountRequestSourceVolume(
                volume_id=volume_id,
                sub_path=sub_path,
            )
        ))

    return volume_file_mounts


def env_vars_callback(ctx: Context, env_vars: List[str]) -> List[str]:
    ctx.store['env_vars'] = []
    for env_var in env_vars:
        try:
            env_key, env_value = env_var.split("=", 1)
            ctx.store['env_vars'].append({
                'key': env_key,
                'value': env_value,
            })
        except ValueError:
            raise ExitException(f'Cannot parse environment variable: {env_var}')

    return env_vars


def git_ref_callback(ctx: Context, git_ref: str) -> str:
    if git_ref:
        git_ref = git_ref.strip()
        if ctx.git_repo and not ctx.git_repo.check_revision_in_remote(git_ref):
            raise ExitException(f'Git commit {git_ref} does not exist in a remote repository.')
        return git_ref

    if ctx.git_repo is None:
        raise ExitException('Run it inside a git-initialized directory or set --git-ref option.')

    return ctx.git_repo.commit_ref


def git_branch_callback(ctx: Context, git_branch: str) -> str:
    if git_branch:
        return git_branch

    if ctx.git_repo is None:
        raise ExitException('Run it inside a git-initialized directory or set --git-branch option.')

    return ctx.git_repo.branch


def git_diff_callback(ctx: Context, git_diff_url: str) -> Optional[str]:
    assert ctx.params['git_ref']
    if ctx.params['ignore_git_diff']:
        return None

    diff_file = None
    if git_diff_url:
        if git_diff_url.startswith('https://') or git_diff_url.startswith('http://'):
            diff_file = tempfile.NamedTemporaryFile(suffix='.patch')
            d = DownloadableFileObject(git_diff_url, os.path.dirname(diff_file.name), os.path.basename(diff_file.name))
            session = requests.Session()
            session.headers = ctx.authenticated_client.auth_header
            d.download(session=session)
            diff_file.seek(0)
    else:
        if ctx.git_repo is None:
            return None

        typer.echo(f'Run experiment with revision {ctx.git_repo.commit_ref[:7]} ({ctx.git_repo.branch})')
        typer.echo(f'Commit: {ctx.git_repo.get_commit_message(ctx.git_repo.commit_ref)}')
        if not ctx.git_repo.is_head:
            typer.echo('Your current revision does not exist in remote repository. '
                       'SavviHub will use latest remote branch revision hash and uncommitted diff.')
        typer.echo('')

        has_diff, diff_status = ctx.git_repo.get_current_diff_status(ctx.git_repo.commit_ref)
        if has_diff:
            typer.echo('Diff to be uploaded: ')

            uncommitted_files = diff_status.get('uncommitted')
            untracked_files = diff_status.get('untracked')

            if uncommitted_files:
                typer.echo('  Changes not committed')
                typer.echo('\n'.join([f'    {x}' for x in uncommitted_files]))
                typer.echo('')
            if untracked_files:
                typer.echo(f'  Untracked files:')
                typer.echo('\n'.join([f'    {x}' for x in untracked_files]))
                typer.echo('')

            answer = inquirer.prompt([inquirer.List(
                'question',
                message='Run experiment with diff?',
                choices=[
                    ('[1] Run experiment with uncommitted and untracked changes.', 1),
                    ('[2] Run experiment with uncommitted changes.', 2),
                    ('[3] Run experiment without any changes.', 3),
                    ('[4] Abort.', 4),
                ],
            )])['question']

            if answer == 1:
                diff_file = ctx.git_repo.get_current_diff_file(ctx.git_repo.commit_ref, with_untracked=True)
            elif answer == 2:
                diff_file = ctx.git_repo.get_current_diff_file(ctx.git_repo.commit_ref, with_untracked=False)
            elif answer == 3:
                pass
            else:
                raise ExitException('Aborted.')

    if diff_file:
        typer.echo('Generating diff patch file...')
        uploaded = Uploader.upload(
            ctx.authenticated_client,
            local_path=diff_file.name,
            volume_id=ctx.project.volume_id,
            remote_path=os.path.basename(diff_file.name),
            progressable=typer.progressbar,
        )
        diff_file_path = uploaded.path
        diff_file.close()

        ctx.store['git_diff_uploaded_path'] = diff_file_path

    return git_diff_url



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import inspect
import pprint
import re  # noqa: F401
import six

from openapi_client.configuration import Configuration


class ResponseKernelImage(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'command': 'str',
        'created_dt': 'datetime',
        'description': 'str',
        'execute_type': 'str',
        'id': 'int',
        'image_pull_policy': 'str',
        'image_url': 'str',
        'immutable_slug': 'str',
        'is_public': 'bool',
        'is_savvihub_managed': 'bool',
        'language': 'str',
        'name': 'str',
        'packages': 'str',
        'processor_type': 'str',
        'updated_dt': 'datetime',
        'working_dir': 'str',
        'workspace_credentials_id': 'int',
        'workspace_id': 'int'
    }

    attribute_map = {
        'command': 'command',
        'created_dt': 'created_dt',
        'description': 'description',
        'execute_type': 'execute_type',
        'id': 'id',
        'image_pull_policy': 'image_pull_policy',
        'image_url': 'image_url',
        'immutable_slug': 'immutable_slug',
        'is_public': 'is_public',
        'is_savvihub_managed': 'is_savvihub_managed',
        'language': 'language',
        'name': 'name',
        'packages': 'packages',
        'processor_type': 'processor_type',
        'updated_dt': 'updated_dt',
        'working_dir': 'working_dir',
        'workspace_credentials_id': 'workspace_credentials_id',
        'workspace_id': 'workspace_id'
    }

    def __init__(self, command=None, created_dt=None, description=None, execute_type=None, id=None, image_pull_policy=None, image_url=None, immutable_slug=None, is_public=None, is_savvihub_managed=None, language=None, name=None, packages=None, processor_type=None, updated_dt=None, working_dir=None, workspace_credentials_id=None, workspace_id=None, local_vars_configuration=None):  # noqa: E501
        """ResponseKernelImage - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._command = None
        self._created_dt = None
        self._description = None
        self._execute_type = None
        self._id = None
        self._image_pull_policy = None
        self._image_url = None
        self._immutable_slug = None
        self._is_public = None
        self._is_savvihub_managed = None
        self._language = None
        self._name = None
        self._packages = None
        self._processor_type = None
        self._updated_dt = None
        self._working_dir = None
        self._workspace_credentials_id = None
        self._workspace_id = None
        self.discriminator = None

        self.command = command
        self.created_dt = created_dt
        self.description = description
        self.execute_type = execute_type
        self.id = id
        self.image_pull_policy = image_pull_policy
        self.image_url = image_url
        self.immutable_slug = immutable_slug
        self.is_public = is_public
        self.is_savvihub_managed = is_savvihub_managed
        self.language = language
        self.name = name
        self.packages = packages
        self.processor_type = processor_type
        self.updated_dt = updated_dt
        self.working_dir = working_dir
        self.workspace_credentials_id = workspace_credentials_id
        self.workspace_id = workspace_id

    @property
    def command(self):
        """Gets the command of this ResponseKernelImage.  # noqa: E501


        :return: The command of this ResponseKernelImage.  # noqa: E501
        :rtype: str
        """
        return self._command

    @command.setter
    def command(self, command):
        """Sets the command of this ResponseKernelImage.


        :param command: The command of this ResponseKernelImage.  # noqa: E501
        :type command: str
        """
        if self.local_vars_configuration.client_side_validation and command is None:  # noqa: E501
            raise ValueError("Invalid value for `command`, must not be `None`")  # noqa: E501

        self._command = command

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseKernelImage.  # noqa: E501


        :return: The created_dt of this ResponseKernelImage.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseKernelImage.


        :param created_dt: The created_dt of this ResponseKernelImage.  # noqa: E501
        :type created_dt: datetime
        """

        self._created_dt = created_dt

    @property
    def description(self):
        """Gets the description of this ResponseKernelImage.  # noqa: E501


        :return: The description of this ResponseKernelImage.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this ResponseKernelImage.


        :param description: The description of this ResponseKernelImage.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def execute_type(self):
        """Gets the execute_type of this ResponseKernelImage.  # noqa: E501


        :return: The execute_type of this ResponseKernelImage.  # noqa: E501
        :rtype: str
        """
        return self._execute_type

    @execute_type.setter
    def execute_type(self, execute_type):
        """Sets the execute_type of this ResponseKernelImage.


        :param execute_type: The execute_type of this ResponseKernelImage.  # noqa: E501
        :type execute_type: str
        """

        self._execute_type = execute_type

    @property
    def id(self):
        """Gets the id of this ResponseKernelImage.  # noqa: E501


        :return: The id of this ResponseKernelImage.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseKernelImage.


        :param id: The id of this ResponseKernelImage.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def image_pull_policy(self):
        """Gets the image_pull_policy of this ResponseKernelImage.  # noqa: E501


        :return: The image_pull_policy of this ResponseKernelImage.  # noqa: E501
        :rtype: str
        """
        return self._image_pull_policy

    @image_pull_policy.setter
    def image_pull_policy(self, image_pull_policy):
        """Sets the image_pull_policy of this ResponseKernelImage.


        :param image_pull_policy: The image_pull_policy of this ResponseKernelImage.  # noqa: E501
        :type image_pull_policy: str
        """
        if self.local_vars_configuration.client_side_validation and image_pull_policy is None:  # noqa: E501
            raise ValueError("Invalid value for `image_pull_policy`, must not be `None`")  # noqa: E501

        self._image_pull_policy = image_pull_policy

    @property
    def image_url(self):
        """Gets the image_url of this ResponseKernelImage.  # noqa: E501


        :return: The image_url of this ResponseKernelImage.  # noqa: E501
        :rtype: str
        """
        return self._image_url

    @image_url.setter
    def image_url(self, image_url):
        """Sets the image_url of this ResponseKernelImage.


        :param image_url: The image_url of this ResponseKernelImage.  # noqa: E501
        :type image_url: str
        """
        if self.local_vars_configuration.client_side_validation and image_url is None:  # noqa: E501
            raise ValueError("Invalid value for `image_url`, must not be `None`")  # noqa: E501

        self._image_url = image_url

    @property
    def immutable_slug(self):
        """Gets the immutable_slug of this ResponseKernelImage.  # noqa: E501


        :return: The immutable_slug of this ResponseKernelImage.  # noqa: E501
        :rtype: str
        """
        return self._immutable_slug

    @immutable_slug.setter
    def immutable_slug(self, immutable_slug):
        """Sets the immutable_slug of this ResponseKernelImage.


        :param immutable_slug: The immutable_slug of this ResponseKernelImage.  # noqa: E501
        :type immutable_slug: str
        """
        if self.local_vars_configuration.client_side_validation and immutable_slug is None:  # noqa: E501
            raise ValueError("Invalid value for `immutable_slug`, must not be `None`")  # noqa: E501

        self._immutable_slug = immutable_slug

    @property
    def is_public(self):
        """Gets the is_public of this ResponseKernelImage.  # noqa: E501


        :return: The is_public of this ResponseKernelImage.  # noqa: E501
        :rtype: bool
        """
        return self._is_public

    @is_public.setter
    def is_public(self, is_public):
        """Sets the is_public of this ResponseKernelImage.


        :param is_public: The is_public of this ResponseKernelImage.  # noqa: E501
        :type is_public: bool
        """
        if self.local_vars_configuration.client_side_validation and is_public is None:  # noqa: E501
            raise ValueError("Invalid value for `is_public`, must not be `None`")  # noqa: E501

        self._is_public = is_public

    @property
    def is_savvihub_managed(self):
        """Gets the is_savvihub_managed of this ResponseKernelImage.  # noqa: E501


        :return: The is_savvihub_managed of this ResponseKernelImage.  # noqa: E501
        :rtype: bool
        """
        return self._is_savvihub_managed

    @is_savvihub_managed.setter
    def is_savvihub_managed(self, is_savvihub_managed):
        """Sets the is_savvihub_managed of this ResponseKernelImage.


        :param is_savvihub_managed: The is_savvihub_managed of this ResponseKernelImage.  # noqa: E501
        :type is_savvihub_managed: bool
        """
        if self.local_vars_configuration.client_side_validation and is_savvihub_managed is None:  # noqa: E501
            raise ValueError("Invalid value for `is_savvihub_managed`, must not be `None`")  # noqa: E501

        self._is_savvihub_managed = is_savvihub_managed

    @property
    def language(self):
        """Gets the language of this ResponseKernelImage.  # noqa: E501


        :return: The language of this ResponseKernelImage.  # noqa: E501
        :rtype: str
        """
        return self._language

    @language.setter
    def language(self, language):
        """Sets the language of this ResponseKernelImage.


        :param language: The language of this ResponseKernelImage.  # noqa: E501
        :type language: str
        """

        self._language = language

    @property
    def name(self):
        """Gets the name of this ResponseKernelImage.  # noqa: E501


        :return: The name of this ResponseKernelImage.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ResponseKernelImage.


        :param name: The name of this ResponseKernelImage.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def packages(self):
        """Gets the packages of this ResponseKernelImage.  # noqa: E501


        :return: The packages of this ResponseKernelImage.  # noqa: E501
        :rtype: str
        """
        return self._packages

    @packages.setter
    def packages(self, packages):
        """Sets the packages of this ResponseKernelImage.


        :param packages: The packages of this ResponseKernelImage.  # noqa: E501
        :type packages: str
        """
        if self.local_vars_configuration.client_side_validation and packages is None:  # noqa: E501
            raise ValueError("Invalid value for `packages`, must not be `None`")  # noqa: E501

        self._packages = packages

    @property
    def processor_type(self):
        """Gets the processor_type of this ResponseKernelImage.  # noqa: E501


        :return: The processor_type of this ResponseKernelImage.  # noqa: E501
        :rtype: str
        """
        return self._processor_type

    @processor_type.setter
    def processor_type(self, processor_type):
        """Sets the processor_type of this ResponseKernelImage.


        :param processor_type: The processor_type of this ResponseKernelImage.  # noqa: E501
        :type processor_type: str
        """

        self._processor_type = processor_type

    @property
    def updated_dt(self):
        """Gets the updated_dt of this ResponseKernelImage.  # noqa: E501


        :return: The updated_dt of this ResponseKernelImage.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this ResponseKernelImage.


        :param updated_dt: The updated_dt of this ResponseKernelImage.  # noqa: E501
        :type updated_dt: datetime
        """

        self._updated_dt = updated_dt

    @property
    def working_dir(self):
        """Gets the working_dir of this ResponseKernelImage.  # noqa: E501


        :return: The working_dir of this ResponseKernelImage.  # noqa: E501
        :rtype: str
        """
        return self._working_dir

    @working_dir.setter
    def working_dir(self, working_dir):
        """Sets the working_dir of this ResponseKernelImage.


        :param working_dir: The working_dir of this ResponseKernelImage.  # noqa: E501
        :type working_dir: str
        """

        self._working_dir = working_dir

    @property
    def workspace_credentials_id(self):
        """Gets the workspace_credentials_id of this ResponseKernelImage.  # noqa: E501


        :return: The workspace_credentials_id of this ResponseKernelImage.  # noqa: E501
        :rtype: int
        """
        return self._workspace_credentials_id

    @workspace_credentials_id.setter
    def workspace_credentials_id(self, workspace_credentials_id):
        """Sets the workspace_credentials_id of this ResponseKernelImage.


        :param workspace_credentials_id: The workspace_credentials_id of this ResponseKernelImage.  # noqa: E501
        :type workspace_credentials_id: int
        """

        self._workspace_credentials_id = workspace_credentials_id

    @property
    def workspace_id(self):
        """Gets the workspace_id of this ResponseKernelImage.  # noqa: E501


        :return: The workspace_id of this ResponseKernelImage.  # noqa: E501
        :rtype: int
        """
        return self._workspace_id

    @workspace_id.setter
    def workspace_id(self, workspace_id):
        """Sets the workspace_id of this ResponseKernelImage.


        :param workspace_id: The workspace_id of this ResponseKernelImage.  # noqa: E501
        :type workspace_id: int
        """
        if self.local_vars_configuration.client_side_validation and workspace_id is None:  # noqa: E501
            raise ValueError("Invalid value for `workspace_id`, must not be `None`")  # noqa: E501

        self._workspace_id = workspace_id

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = inspect.getargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseKernelImage):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseKernelImage):
            return True

        return self.to_dict() != other.to_dict()



from savvihub.exceptions import SavviHubException


class ExitException(SavviHubException):
    """ Usage: raise ExitError('error message') """


class InvalidGitRepository(SavviHubException):
    pass





import inspect
from functools import wraps

from openapi_client import *
from savvihub.api.exceptions import convert_to_savvihub_exception, NotFoundAPIException
from savvihub.common.constants import API_HOST


def raise_savvihub_exception(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except ApiException as e:
            raise convert_to_savvihub_exception(e)

    return wrapped


def decorate_all_methods(method_decorator):
    def class_decorator(cls):
        for attr in inspect.classify_class_attrs(cls):
            if attr.kind == 'method' and issubclass(attr.defining_class, cls.__bases__[0]):
                setattr(cls, attr.name, method_decorator(attr[3]))
        return cls
    return class_decorator


@decorate_all_methods(raise_savvihub_exception)
class OpenAPIClientWithSavviHubException(APIV1Api):
    pass


class SavviHubClient:
    def __init__(self, *, auth_header=None, api_host=API_HOST):
        api_configuration = Configuration(host=api_host)
        # disable verify ssl (https://github.com/urllib3/urllib3/issues/1682#issuecomment-533311857)
        api_configuration.verify_ssl = False
        import urllib3
        urllib3.disable_warnings()

        kwargs = {
            'configuration': api_configuration,
        }

        self.auth_header = {}
        if auth_header:
            self.auth_header = auth_header
            for header_name, header_value in auth_header.items():
                kwargs['header_name'], kwargs['header_value'] = header_name, header_value

        self.client = OpenAPIClientWithSavviHubException(ApiClient(**kwargs))

    def signup_for_test_only(self, email, username, name, password, invitation_token) -> ResponseUserWithTokenResponse:
        return self.client.sign_up_api(
            sign_up_api_payload=SignUpAPIPayload(
                email=email,
                username=username,
                name=name,
                password=password,
                invitation_token=invitation_token,
            ),
        )

    def signin_confirm_for_test_only(self, cli_token):
        return self.client.sign_in_cli_confirm_api(
            sign_in_cli_confirm_api_payload=SignInCliConfirmAPIPayload(
                cli_token=cli_token,
            ),
        )

    def get_my_info(self) -> ResponseMyUser:
        return self.client.get_my_user_info_api()

    def region_list(self) -> RegionRegionListResponse:
        return self.client.region_list_api()

    def verify_access_token(self):
        self.client.access_token_verify_api()

    def experiment_id_read(self, experiment_id) -> ResponseExperimentInfo:
        return self.client.experiment_read_by_idapi(experiment_id=experiment_id)

    def project_read(self, workspace_name, project_name) -> ResponseProjectInfo:
        return self.client.project_read_api(workspace_name=workspace_name, project_name=project_name)

    def signin_cli_token(self) -> AccountSignInCliTokenResponse:
        return self.client.sign_in_cli_token_api()

    def check_signin(self, cli_token) -> AccountSignInCliCheckResponse:
        return self.client.sign_in_cli_check_api(cli_token=cli_token)

    def volume_read(self, volume_id) -> ResponseVolume:
        return self.client.volume_read_api(volume_id=volume_id)

    def volume_file_list(
        self, volume_id, snapshot='latest', path='', recursive=False, need_download_url=False,
    ) -> VolumeVolumeFileListResponse:
        return self.client.volume_file_list_api(
            volume_id=volume_id,
            path=path,
            recursive=recursive,
            snapshot=snapshot,
            need_download_url=need_download_url,
        )

    def volume_file_read(self, volume_id, path, snapshot) -> ResponseFileMetadata:
        return self.client.volume_file_read_api(volume_id=volume_id, path=path, snapshot=snapshot)

    def volume_file_copy(self, volume_id, source_path, source_snapshot, dest_path, recursive=False) -> VolumeVolumeFileCopyResponse:
        return self.client.volume_file_copy_api(
            volume_id=volume_id,
            volume_file_copy_api_payload=VolumeFileCopyAPIPayload(
                source_path=source_path,
                source_snapshot=source_snapshot,
                dest_path=dest_path,
                recursive=recursive,
            ),
        )

    def volume_file_delete(self, volume_id, path, recursive=False) -> VolumeVolumeFileDeleteResponse:
        return self.client.volume_file_delete_api(volume_id=volume_id, path=path, recursive=recursive)

    def volume_file_create(self, volume_id, path, is_dir) -> ResponseFileMetadata:
        return self.client.volume_file_create_api(
            volume_id=volume_id,
            volume_file_create_api_payload=VolumeFileCreateAPIPayload(
                path=path,
                is_dir=is_dir,
            ),
        )

    def volume_file_uploaded(self, volume_id, path) -> ResponseFileMetadata:
        return self.client.volume_file_uploaded_api(
            volume_id=volume_id,
            path=path,
        )

    def snapshot_read(self, volume_id, snapshot_name) -> ResponseSnapshot:
        return self.client.snapshot_read_api(volume_id=volume_id, snapshot_name=snapshot_name)

    def experiment_read(self, workspace, project, experiment_number_or_name) -> ResponseExperimentInfo:
        return self.client.experiment_read_api(
            workspace_name=workspace,
            project_name=project,
            experiment=experiment_number_or_name,
        )

    def experiment_list(self, workspace, project) -> ExperimentExperimentListResponse:
        return self.client.experiment_list_by_project_api(
            workspace_name=workspace,
            project_name=project,
            order_field='number',
            order_direction='desc',
        )

    def experiment_log(self, workspace, project, experiment_number_or_name, **kwargs) -> ExperimentExperimentLogResponse:
        return self.client.experiment_log_api(
            workspace_name=workspace,
            project_name=project,
            experiment=experiment_number_or_name,
            **kwargs,
        )

    def experiment_create(
        self, workspace, project, cluster_name, image_url, resource_spec_id, resource_spec,
        start_command, env_vars, volume_mounts,
    ) -> ResponseExperimentInfo:
        return self.client.experiment_create_api(
            workspace_name=workspace,
            project_name=project,
            experiment_create_api_payload=ExperimentCreateAPIPayload(
                cluster_name=cluster_name,
                image_url=image_url,
                resource_spec_id=resource_spec_id,
                resource_spec=resource_spec,
                env_vars=env_vars,
                start_command=start_command,
                volumes=volume_mounts,
            ),
        )

    def experiment_plots_metrics_update(self, experiment_id, metrics):
        self.client.cli_experiment_plots_metrics_update_api(
            experiment_id=experiment_id,
            cli_experiment_plots_metrics_update_api_payload=CliExperimentPlotsMetricsUpdateAPIPayload(
                metrics=metrics,
            ),
        )

    def experiment_plots_files_update(self, experiment_id, files, type):
        self.client.cli_experiment_plots_files_update_api(
            experiment_id=experiment_id,
            cli_experiment_plots_files_update_api_payload=CliExperimentPlotsFilesUpdateAPIPayload(
                files=files,
                type=type,
            )
        )

    def kernel_image_list(self, workspace) -> KernelKernelImageListResponse:
        return self.client.kernel_image_list_api(workspace_name=workspace)

    def kernel_resource_list(self, workspace) -> KernelKernelResourceSpecListResponse:
        return self.client.kernel_resource_spec_list_api(workspace_name=workspace)

    def workspace_list(self) -> WorkspaceWorkspaceListResponse:
        return self.client.workspace_list_api()

    def workspace_read(self, workspace) -> ResponseWorkspace:
        return self.client.workspace_read_api(workspace_name=workspace)

    def workspace_create(self, workspace_name, region) -> ResponseWorkspace:
        return self.client.workspace_create_api(
            workspace_create_api_payload=WorkspaceCreateAPIPayload(
                name=workspace_name,
                region=region,
            ),
        )

    def project_list(self, workspace_name) -> ResponseProjectListResponse:
        return self.client.project_list_api(workspace_name=workspace_name)

    def project_github_create(self, workspace, project, github_owner, github_repo) -> ResponseProject:
        return self.client.project_git_hub_create_api(
            workspace_name=workspace,
            project_git_hub_create_api_payload=ProjectGitHubCreateAPIPayload(
                name=project,
                github_owner=github_owner,
                github_repo=github_repo,
            ),
        )

    def public_dataset_list(self) -> ResponseDatasetInfoList:
        return self.client.datasets_public_list_api()

    def dataset_list(self, workspace) -> ResponseDatasetInfoList:
        return self.client.dataset_list_api(workspace_name=workspace)

    def dataset_read(self, workspace, dataset) -> ResponseDatasetInfo:
        try:
            return self.client.dataset_read_api(workspace_name=workspace, dataset_name=dataset)
        except NotFoundAPIException:
            raise NotFoundAPIException(f'Dataset `{dataset}` is not in the workspace `{workspace}`.')

    def dataset_create(self, workspace, name, is_public, description) -> ResponseDatasetInfo:
        return self.client.savvi_hub_dataset_create_api(
            workspace_name=workspace,
            savvi_hub_dataset_create_api_payload=SavviHubDatasetCreateAPIPayload(
                name=name,
                is_public=is_public,
                description=description,
            ),
        )

    def dataset_gs_create(self, workspace, name, is_public, description, gs_path) -> ResponseDatasetInfo:
        return self.client.g_s_dataset_create_api(
            workspace_name=workspace,
            gs_dataset_create_api_payload=GSDatasetCreateAPIPayload(
                name=name,
                is_public=is_public,
                description=description,
                gs_path=gs_path,
            ),
        )

    def dataset_s3_create(self, workspace, name, is_public, description, s3_path, aws_role_arn) -> ResponseDatasetInfo:
        return self.client.s3_dataset_create_api(
            workspace_name=workspace,
            s3_dataset_create_api_payload=S3DatasetCreateAPIPayload(
                name=name,
                is_public=is_public,
                description=description,
                s3_path=s3_path,
                aws_role_arn=aws_role_arn,
            ),
        )

    def cluster_list(self, workspace_name) -> ClusterAllClusterListResponse:
        return self.client.all_cluster_list_api(workspace_name=workspace_name)

    def cluster_delete(self, workspace_name, cluster_name):
        self.client.custom_cluster_delete_api(workspace_name=workspace_name, cluster_name=cluster_name)

    def cluster_rename(self, workspace_name, cluster_name, new_name) -> ResponseKernelCluster:
        return self.client.custom_cluster_update_api(
            workspace_name=workspace_name,
            cluster_name=cluster_name,
            custom_cluster_update_api_payload=CustomClusterUpdateAPIPayload(
                name=new_name,
            ),
        )

    def cluster_node_list(self, workspace_name, cluster_name) -> ClusterCustomClusterNodeListResponse:
        return self.client.custom_cluster_node_list_api(
            workspace_name=workspace_name,
            cluster_name=cluster_name
        )



import os
from pathlib import Path

WEB_HOST = os.environ.get('SAVVIHUB_WEB_HOST', 'https://savvihub.com')

CUR_DIR = os.getcwd()
DEFAULT_SAVVI_DIR = os.path.join(str(Path.home()), '.savvihub')
DEFAULT_CONFIG_PATH = os.path.join(DEFAULT_SAVVI_DIR, 'config')
PROJECT_ROOT = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

DATASET_SOURCE_TYPE_SAVVIHUB_S3 = 'savvihub-s3'
DATASET_SOURCE_TYPE_AWS_S3 = 's3'
DATASET_SOURCE_TYPE_GCP_GS = 'gcp-gs'

DATASET_PATH_PARSE_SCHEME_GS = "gs"
DATASET_PATH_PARSE_SCHEME_S3 = "s3"



import os
import sys
project_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
sys.path.append(project_root)

import time

import requests
import typer

from savvihub.api.exceptions import NotFoundAPIException, InvalidTokenAPIException
from savvihub.api.savvihub import SavviHubClient
from savvihub.cli.commands.cluster import cluster_app
from savvihub.cli.commands.dataset import dataset_app
from savvihub.cli.commands.experiment import experiment_app
from savvihub.cli.commands.volume import volume_app
from savvihub.cli.config_loader import GlobalConfigLoader
from savvihub.cli.constants import WEB_HOST
from savvihub.cli.exceptions import ExitException
from savvihub.cli.typer import Typer, Context
from savvihub.cli.utils import get_default_workspace
from savvihub.common.constants import API_HOST, DEBUG


app = Typer()
app.add_typer(experiment_app, name='experiment')
app.add_typer(dataset_app, name='dataset')
app.add_typer(cluster_app, name='cluster')
app.add_typer(volume_app, name='volume', hidden=True)
__version__ = '0.0.99.dev5'


def version_callback(value: bool):
    if value:
        typer.echo(f'SavviHub CLI Version: {__version__}')
        sys.exit(0)


@app.callback()
def main(
    version: bool = typer.Option(None, '--version', callback=version_callback,
                                 help='Print the current SavviHub CLI version.'),
):
    """
    SavviHub Command Line Interface (CLI)
    """
    if DEBUG:
        typer.echo(f'SavviHub CLI Version: {__version__}')


@app.command(hidden=True)
def ping():
    """
    Ping to the SavviHub server
    """
    res = requests.get(API_HOST + '/api/v1/ping')
    typer.echo(f'Response code: {res.status_code}, Response text: {res.text}')


@app.command()
def login(
    token: str = typer.Option(None, '-t', '--token'),
):
    """
    Initialize SavviHub Command Line Interface (CLI)
    """

    def update_token():
        client = SavviHubClient()
        cli_token = client.signin_cli_token().cli_token
        typer.echo(f'Please grant CLI access from the URL below.\n'
                   f'{WEB_HOST}/cli/grant-access?token={cli_token}')
        typer.echo('Waiting...\n')

        start_time = time.time()
        while True:
            if time.time() - start_time >= 160:
                raise ExitException('Login timeout. Please try again.')

            check_signin_response = client.check_signin(cli_token)
            if not check_signin_response.signin_success:
                time.sleep(3)
                continue

            return check_signin_response.access_token

    if token is None:
        token = update_token()

    client = SavviHubClient(auth_header={'Authorization': f'Token {token}'})
    me = client.get_my_info()
    if me:
        typer.echo(f'Hello, {me.username}!\n')
    else:
        typer.echo('Token expired or invalid token.')
        token = update_token()

    config = GlobalConfigLoader()
    config.token = token
    config.workspace = get_default_workspace(client).name
    typer.echo(f'Successfully configured in {config.filename}')


@app.command()
def status(
    ctx: Context,
):
    """
    Show current user and default workspace
    """
    if ctx.global_config.token is None:
        raise ExitException('Not logged in. Please run `savvihub login`.')

    try:
        user = ctx.authenticated_client.get_my_info()
    except (NotFoundAPIException, InvalidTokenAPIException):
        raise ExitException('Token expired. Please run `savvihub login`.')

    typer.echo(f'Username: {user.username}\n'
               f'Email: {user.email}')

    if ctx.global_config.workspace is None:
        typer.echo('Default workspace is not set. Please run `savvihub set-default-workspace`.')
    else:
        typer.echo(f'Default workspace: {ctx.global_config.workspace}')


@app.command(user_required=True)
def set_default_workspace(
    ctx: Context,
):
    """
    Set default workspace
    """
    workspace = get_default_workspace(ctx.authenticated_client)
    ctx.global_config.workspace = workspace.name


if __name__ == '__main__':
    app()
